@@ -1,107 +1,209 @@
 //! Validation implementations and helper types.
 
 use crate::{ArchivedDynMetadata, RegisteredImpl, IMPL_REGISTRY};
-use bytecheck::{CheckBytes, Unreachable};
+use bytecheck::CheckBytes;
 #[cfg(feature = "vtable_cache")]
-use core::sync::atomic::{AtomicU64, Ordering};
-use core::{alloc::Layout, any::TypeId, fmt, marker::PhantomData};
+use core::sync::atomic::Ordering;
+use core::{
+    alloc::Layout,
+    any::{Any, TypeId},
+    convert::Infallible,
+    fmt,
+    marker::PhantomData,
+    ptr,
+};
 use rkyv::{
-    offset_of,
-    validation::{ArchiveBoundsContext, ArchiveMemoryContext, SharedArchiveContext},
-    Fallible,
+    from_archived,
+    validation::{ArchiveContext, SharedContext},
+    Archived, Fallible,
 };
 use rkyv_typename::TypeName;
 use std::{collections::HashMap, error::Error};
 
 /// A context that's object safe and suitable for checking most types.
 pub trait DynContext {
-    /// Checks the given parts of a relative pointer for bounds issues.
+    /// Checks that a relative pointer points to an address within the archive.
+    ///
+    /// See [`bounds_check_ptr`] for more information.
     ///
     /// # Safety
     ///
-    /// The caller must guarantee that the base pointer is inside the archive for this context.
-    unsafe fn check_rel_ptr_dyn(
+    /// - `base` must be inside the archive this valiator was created for.
+    ///
+    /// [`bounds_check_ptr`]: rkyv::validation::ArchiveContext::bounds_check_ptr
+    unsafe fn bounds_check_ptr_dyn(
         &mut self,
         base: *const u8,
         offset: isize,
     ) -> Result<*const u8, Box<dyn Error>>;
 
-    /// Checks the given memory block for bounds issues.
+    /// Checks that a given pointer can be dereferenced.
+    ///
+    /// See [`bounds_check_layout`] for more information.
     ///
     /// # Safety
     ///
-    /// The caller must guarantee that the pointer is inside the archive for this context.
-    unsafe fn bounds_check_ptr_dyn(
+    /// - `data_address` must be inside the archive this validator was created for.
+    /// - `layout` must be the layout for the given pointer.
+    ///
+    /// [`bounds_check_layout`]: rkyv::validation::ArchiveContext::bounds_check_layout
+    unsafe fn bounds_check_layout_dyn(
         &mut self,
-        ptr: *const u8,
+        data_address: *const u8,
         layout: &Layout,
     ) -> Result<(), Box<dyn Error>>;
 
-    /// Claims `count` bytes located `offset` bytes away from `base`.
+    /// Checks that the given data address and layout is located completely within the subtree
+    /// range.
+    ///
+    /// See [`bounds_check_subtree_ptr_layout`] for more information.
     ///
     /// # Safety
     ///
-    /// The caller must guarantee that `base` is inside the archive this context was created for.
-    unsafe fn claim_bytes_dyn(
+    /// - `data_address` must be inside the archive this validator was created for.
+    ///
+    /// [`bounds_check_subtree_ptr_layout`]: rkyv::validation::ArchiveContext::bounds_check_subtree_ptr_layout
+    unsafe fn bounds_check_subtree_ptr_layout_dyn(
         &mut self,
-        start: *const u8,
-        len: usize,
+        data_address: *const u8,
+        layout: &Layout,
     ) -> Result<(), Box<dyn Error>>;
 
-    /// Claims `count` shared bytes located `offset` bytes away from `base`.
+    /// Pushes a new subtree range onto the validator and starts validating it.
     ///
-    /// Returns whether the bytes need to be checked.
+    /// See [`push_prefix_subtree_range`] for more information.
     ///
     /// # Safety
     ///
-    /// The caller must guarantee that `base` is inside the archive this context was created for.
-    unsafe fn claim_shared_bytes_dyn(
+    /// `root` and `end` must be located inside the archive.
+    ///
+    /// [`push_prefix_subtree_range`]: rkyv::validation::ArchiveContext::push_prefix_subtree_range
+    unsafe fn push_prefix_subtree_range_dyn(
+        &mut self,
+        root: *const u8,
+        end: *const u8,
+    ) -> Result<Box<dyn Any>, Box<dyn Error>>;
+
+    /// Pops the given range, restoring the original state with the pushed range removed.
+    ///
+    /// See [`pop_prefix_range`] for more information.
+    ///
+    /// [`pop_prefix_range`]: rkyv::validation::ArchiveContext::pop_prefix_range
+    fn pop_prefix_range_dyn(&mut self, range: Box<dyn Any>) -> Result<(), Box<dyn Error>>;
+
+    /// Pushes a new subtree range onto the validator and starts validating it.
+    ///
+    /// See [`push_suffix_subtree_range`] for more information.
+    ///
+    /// # Safety
+    ///
+    /// `start` and `root` must be located inside the archive.
+    ///
+    /// [`push_suffix_subtree_range`]: rkyv::validation::ArchiveContext::push_suffix_subtree_range
+    unsafe fn push_suffix_subtree_range_dyn(
         &mut self,
         start: *const u8,
-        len: usize,
+        root: *const u8,
+    ) -> Result<Box<dyn Any>, Box<dyn Error>>;
+
+    /// Finishes the given range, restoring the original state with the pushed range removed.
+    ///
+    /// See [`pop_suffix_range`] for more information.
+    ///
+    /// [`pop_suffix_range`]: rkyv::validation::ArchiveContext::pop_suffix_range
+    fn pop_suffix_range_dyn(&mut self, range: Box<dyn Any>) -> Result<(), Box<dyn Error>>;
+
+    /// Verifies that all outstanding claims have been returned.
+    ///
+    /// See [`finish`] for more information.
+    ///
+    /// [`finish`]: rkyv::validation::ArchiveContext::finish
+    fn finish_dyn(&mut self) -> Result<(), Box<dyn Error>>;
+
+    /// Registers the given `ptr` as a shared pointer with the given type.
+    ///
+    /// See [`register_shared_ptr`] for more information.
+    ///
+    /// [`register_shared_ptr`]: rkyv::validation::SharedContext::register_shared_ptr
+    fn register_shared_ptr_dyn(
+        &mut self,
+        ptr: *const u8,
         type_id: TypeId,
     ) -> Result<bool, Box<dyn Error>>;
 }
 
-impl<C: ArchiveBoundsContext + ArchiveMemoryContext + SharedArchiveContext + ?Sized> DynContext
-    for C
+impl<C> DynContext for C
 where
+    C: ArchiveContext + SharedContext + ?Sized,
     C::Error: Error,
 {
-    unsafe fn check_rel_ptr_dyn(
+    unsafe fn bounds_check_ptr_dyn(
         &mut self,
         base: *const u8,
         offset: isize,
     ) -> Result<*const u8, Box<dyn Error>> {
-        self.check_rel_ptr(base, offset)
+        self.bounds_check_ptr(base, offset)
             .map_err(|e| Box::new(e) as Box<dyn Error>)
     }
 
-    unsafe fn bounds_check_ptr_dyn(
+    unsafe fn bounds_check_layout_dyn(
         &mut self,
         ptr: *const u8,
         layout: &Layout,
     ) -> Result<(), Box<dyn Error>> {
-        self.bounds_check_ptr(ptr, layout)
+        self.bounds_check_layout(ptr, layout)
             .map_err(|e| Box::new(e) as Box<dyn Error>)
     }
 
-    unsafe fn claim_bytes_dyn(
+    unsafe fn bounds_check_subtree_ptr_layout_dyn(
         &mut self,
-        start: *const u8,
-        len: usize,
+        data_address: *const u8,
+        layout: &Layout,
     ) -> Result<(), Box<dyn Error>> {
-        self.claim_bytes(start, len)
+        self.bounds_check_subtree_ptr_layout(data_address, layout)
             .map_err(|e| Box::new(e) as Box<dyn Error>)
     }
 
-    unsafe fn claim_shared_bytes_dyn(
+    unsafe fn push_prefix_subtree_range_dyn(
+        &mut self,
+        root: *const u8,
+        end: *const u8,
+    ) -> Result<Box<dyn Any>, Box<dyn Error>> {
+        self.push_prefix_subtree_range(root, end)
+            .map(|r| Box::new(r) as Box<dyn Any>)
+            .map_err(|e| Box::new(e) as Box<dyn Error>)
+    }
+
+    fn pop_prefix_range_dyn(&mut self, range: Box<dyn Any>) -> Result<(), Box<dyn Error>> {
+        self.pop_prefix_range(*range.downcast().unwrap())
+            .map_err(|e| Box::new(e) as Box<dyn Error>)
+    }
+
+    unsafe fn push_suffix_subtree_range_dyn(
         &mut self,
         start: *const u8,
-        len: usize,
+        root: *const u8,
+    ) -> Result<Box<dyn Any>, Box<dyn Error>> {
+        self.push_suffix_subtree_range(start, root)
+            .map(|r| Box::new(r) as Box<dyn Any>)
+            .map_err(|e| Box::new(e) as Box<dyn Error>)
+    }
+
+    fn pop_suffix_range_dyn(&mut self, range: Box<dyn Any>) -> Result<(), Box<dyn Error>> {
+        self.pop_suffix_range(*range.downcast().unwrap())
+            .map_err(|e| Box::new(e) as Box<dyn Error>)
+    }
+
+    fn finish_dyn(&mut self) -> Result<(), Box<dyn Error>> {
+        self.finish().map_err(|e| Box::new(e) as Box<dyn Error>)
+    }
+
+    fn register_shared_ptr_dyn(
+        &mut self,
+        ptr: *const u8,
         type_id: TypeId,
     ) -> Result<bool, Box<dyn Error>> {
-        self.claim_shared_bytes(start, len, type_id)
+        self.register_shared_ptr(ptr, type_id)
             .map_err(|e| Box::new(e) as Box<dyn Error>)
     }
 }
@@ -110,38 +212,70 @@ impl Fallible for (dyn DynContext + '_) {
     type Error = Box<dyn Error>;
 }
 
-impl ArchiveBoundsContext for (dyn DynContext + '_) {
-    unsafe fn check_rel_ptr(
+impl ArchiveContext for (dyn DynContext + '_) {
+    type PrefixRange = Box<dyn Any>;
+    type SuffixRange = Box<dyn Any>;
+
+    unsafe fn bounds_check_ptr(
         &mut self,
         base: *const u8,
         offset: isize,
     ) -> Result<*const u8, Self::Error> {
-        self.check_rel_ptr_dyn(base, offset)
+        self.bounds_check_ptr_dyn(base, offset)
     }
 
-    unsafe fn bounds_check_ptr(
+    unsafe fn bounds_check_layout(
         &mut self,
-        ptr: *const u8,
+        data_address: *const u8,
         layout: &Layout,
     ) -> Result<(), Self::Error> {
-        self.bounds_check_ptr_dyn(ptr, layout)
+        self.bounds_check_layout_dyn(data_address, layout)
     }
-}
 
-impl ArchiveMemoryContext for (dyn DynContext + '_) {
-    unsafe fn claim_bytes(&mut self, start: *const u8, len: usize) -> Result<(), Self::Error> {
-        self.claim_bytes_dyn(start, len)
+    unsafe fn bounds_check_subtree_ptr_layout(
+        &mut self,
+        data_address: *const u8,
+        layout: &Layout,
+    ) -> Result<(), Self::Error> {
+        self.bounds_check_subtree_ptr_layout_dyn(data_address, layout)
     }
-}
 
-impl SharedArchiveContext for (dyn DynContext + '_) {
-    unsafe fn claim_shared_bytes(
+    unsafe fn push_prefix_subtree_range(
+        &mut self,
+        root: *const u8,
+        end: *const u8,
+    ) -> Result<Self::PrefixRange, Self::Error> {
+        self.push_prefix_subtree_range_dyn(root, end)
+    }
+
+    fn pop_prefix_range(&mut self, range: Self::PrefixRange) -> Result<(), Self::Error> {
+        self.pop_prefix_range_dyn(range)
+    }
+
+    unsafe fn push_suffix_subtree_range(
         &mut self,
         start: *const u8,
-        len: usize,
+        root: *const u8,
+    ) -> Result<Self::SuffixRange, Self::Error> {
+        self.push_suffix_subtree_range_dyn(start, root)
+    }
+
+    fn pop_suffix_range(&mut self, range: Self::SuffixRange) -> Result<(), Self::Error> {
+        self.pop_suffix_range_dyn(range)
+    }
+
+    fn finish(&mut self) -> Result<(), Self::Error> {
+        self.finish_dyn()
+    }
+}
+
+impl SharedContext for (dyn DynContext + '_) {
+    fn register_shared_ptr(
+        &mut self,
+        ptr: *const u8,
         type_id: TypeId,
     ) -> Result<bool, Box<dyn Error>> {
-        self.claim_shared_bytes_dyn(start, len, type_id)
+        self.register_shared_ptr_dyn(ptr, type_id)
     }
 }
 
@@ -160,6 +294,7 @@ impl Error for CheckBytesUnimplemented {}
 type CheckBytesDyn = unsafe fn(*const u8, &mut dyn DynContext) -> Result<(), Box<dyn Error>>;
 
 // This is the fallback function that gets called if the archived type doesn't implement CheckBytes.
+#[inline]
 unsafe fn check_bytes_dyn_unimplemented(
     _bytes: *const u8,
     _context: &mut dyn DynContext,
@@ -178,8 +313,10 @@ impl<T: ?Sized> NotCheckBytesDyn for T {}
 pub struct IsCheckBytesDyn<T: ?Sized>(PhantomData<T>);
 
 impl<T: for<'a> CheckBytes<dyn DynContext + 'a>> IsCheckBytesDyn<T> {
+    #[doc(hidden)]
     pub const CHECK_BYTES_DYN: CheckBytesDyn = Self::check_bytes_dyn;
 
+    #[inline]
     unsafe fn check_bytes_dyn(
         bytes: *const u8,
         context: &mut dyn DynContext,
@@ -243,8 +380,8 @@ impl fmt::Display for DynMetadataError {
 
 impl Error for DynMetadataError {}
 
-impl From<Unreachable> for DynMetadataError {
-    fn from(_: Unreachable) -> Self {
+impl From<Infallible> for DynMetadataError {
+    fn from(_: Infallible) -> Self {
         unsafe { core::hint::unreachable_unchecked() }
     }
 }
@@ -256,17 +393,19 @@ impl<T: TypeName + ?Sized, C: ?Sized> CheckBytes<C> for ArchivedDynMetadata<T> {
         value: *const Self,
         context: &mut C,
     ) -> Result<&'a Self, Self::Error> {
-        let bytes = value.cast::<u8>();
-
-        let type_id = *u64::check_bytes(bytes.add(offset_of!(Self, type_id)).cast(), context)?;
-        PhantomData::<T>::check_bytes(bytes.add(offset_of!(Self, phantom)).cast(), context)?;
+        let type_id = from_archived!(*Archived::<u64>::check_bytes(
+            ptr::addr_of!((*value).type_id),
+            context,
+        )?);
+        PhantomData::<T>::check_bytes(ptr::addr_of!((*value).phantom), context)?;
         if let Some(impl_data) = IMPL_REGISTRY.get::<T>(type_id) {
-            let cached_vtable_ptr = bytes.add(offset_of!(Self, cached_vtable));
+            let cached_vtable_ptr = ptr::addr_of!((*value).cached_vtable);
             #[cfg(feature = "vtable_cache")]
             let cached_vtable =
-                AtomicU64::check_bytes(cached_vtable_ptr.cast(), context)?.load(Ordering::Relaxed);
+                CheckBytes::check_bytes(cached_vtable_ptr, context)?.load(Ordering::Relaxed);
             #[cfg(not(feature = "vtable_cache"))]
-            let cached_vtable = *u64::check_bytes(cached_vtable_ptr.cast(), context)?;
+            let cached_vtable =
+                from_archived!(*Archived::<u64>::check_bytes(cached_vtable_ptr, context)?);
             if cached_vtable == 0 || cached_vtable as usize == impl_data.vtable {
                 Ok(&*value)
             } else {
@@ -322,6 +461,7 @@ pub struct CheckBytesEntry {
 }
 
 impl CheckBytesEntry {
+    #[doc(hidden)]
     pub fn new<TY: RegisteredImpl<TR>, TR: ?Sized>(check_bytes_dyn: CheckBytesDyn) -> Self {
         Self {
             vtable: <TY as RegisteredImpl<TR>>::vtable(),
@@ -355,6 +495,7 @@ impl CheckBytesRegistry {
         debug_assert!(old_value.is_none(), "vtable conflict, a trait implementation was likely added twice (but it's possible there was a hash collision)");
     }
 
+    #[doc(hidden)]
     pub fn get(&self, vtable: usize) -> Option<&ImplValidation> {
         self.vtable_to_check_bytes.get(&vtable)
     }
